#!/usr/bin/env python3
"""Provides infastructure for font selection in Rho+."""

from importlib.resources import path


def mpl_add_fonts():
    """Configures Matplotlib to add Source Sans 3."""
    import matplotlib.font_manager as fm

    fonts = ["SourceSans3-Regular.ttf", "SourceSans3-Bold.ttf"]
    for f in fonts:
        with path("rho_plus.data.fonts", f) as ttf:
            fm.fontManager.addfont(ttf)



#!/usr/bin/env python3
"""Colors used across themes."""

LIGHT_COLORS = [
    "#1846b3",
    "#355e00",
    "#bb8600",
    "#dd2800",
    "#c510d2",
    "#9470fe",
    "#007dc2",
    "#00aeb3",
    "#008a47",
    "#942e00",
    "#f3398e",
    "#9e0077",
]

DARK_COLORS = [
    "#00a0ec",
    "#00bc70",
    "#deca00",
    "#ff7300",
    "#d83990",
    "#7555d3",
    "#8ac7ff",
    "#00f0ff",
    "#387200",
    "#aa2e00",
    "#ff7dc6",
    "#e960ff",
]


# the neutrals are from https://elastic.github.io/eui/#/theming/colors
# the naming is from a light-scheme perspective

LIGHT_SHADES = [
    # empty shade: primary background
    "#FFFFFF",
    # lightest shade: secondary background
    "#F1F4FA",
    # light shade: borders and dividers
    "#D3DAE6",
    # medium shade: subdued text
    "#98A2B3",
    # dark shade: secondary foreground
    "#69707D",
    # darkest shade: primary foreground, text
    "#343741",
]

DARK_SHADES = [
    # empty shade: primary background
    "#1D1E24",
    # lightest shade: secondary background
    "#25262E",
    # light shade: borders and dividers
    "#343741",
    # medium shade: subdued text
    "#535966",
    # dark shade: secondary foreground
    "#98A2B3",
    # darkest shade: primary foreground, text
    "#D4DAE5",
]



"""Defines a common interface to color palettes used in rho_plus."""

from __future__ import annotations
import abc
from typing import List
import matplotlib.colors as mpl_colors


class SequentialPaletteMixin(abc.ABC):
    """Mixin for sequential palettes that handles conversion to different formats."""

    @abc.abstractmethod
    def name(self) -> str:
        """The name of the palette."""
        raise NotImplementedError("Subclasses must define a name")

    @abc.abstractmethod
    def hex_colors(self) -> List[str]:
        """Returns a list of hex colors to interpolate between."""
        raise NotImplementedError("Subclasses must implement method hex_colors")

    def rev(self) -> SequentialPaletteMixin:
        """Returns a reversed version of the color palette."""
        return ReversedPalette(self)

    def as_mpl_cmap(self) -> mpl_colors.Colormap:
        """Returns a Matplotlib colormap."""
        return mpl_colors.LinearSegmentedColormap.from_list(
            self.name(), self.hex_colors(), N=len(self.hex_colors())
        )


class ReversedPalette(SequentialPaletteMixin):
    """Reversed color palette."""

    def __init__(self, rev_palette: SequentialPaletteMixin):
        self.rev_palette = rev_palette

    def name(self) -> str:
        return self.rev_palette.name() + "_r"

    def hex_colors(self) -> List[str]:
        return self.rev_palette.hex_colors()[::-1]


class SequentialPalette(SequentialPaletteMixin):
    def __init__(self, name: str, colors: List[str]):
        self._name = name
        self._colors = colors

    @classmethod
    def from_mpl_cmap(cls, cmap: mpl_colors.Colormap):
        """Initializes from a Matplotlib cmap object."""
        return cls(cmap.name, [mpl_colors.to_hex(rgb) for rgb in cmap(range(cmap.N))])

    def name(self) -> str:
        return self._name

    def hex_colors(self) -> List[str]:
        return self._colors



#!/usr/bin/env python3
"""General utilities."""

import re
from typing import Iterable, Union
import numpy as np


def spread(x, dmin):
    x = np.array(x, copy=False).astype(np.float64)
    if len(x) <= 1:
        return x
    
    # we need to work on sorted x, but we also need to remember the indices so we can adjust the right ones
    sort_inds = np.argsort(x)
    x_sort = x.copy()[sort_inds]

    dmin = np.broadcast_to(dmin, x.shape)[sort_inds]

    # Minimizing the Total Movement for Movement to Independence Problem on a Line
    # Ghadiri, Yazdanbod 2016
    # https://www.researchgate.net/publication/304641457_Minimizing_the_Total_Movement_for_Movement_to_Independence_Problem_on_a_Line
    # this function does things a little differently, which I think results in clearer code
    
    slack = np.zeros_like(x_sort, dtype=np.float64)
    adj = np.zeros_like(x_sort, dtype=np.float64)
    for i in range(1, len(x_sort)):
        # previous element's new location is x_sort[i-1] + adj[i-1]
        # move to max(x_sort[i], prev + dmin)

        # dmin = max(dmin[i], dmin[i-1])

        local_dmin = max(dmin[i], dmin[i-1])
        prev = x_sort[i-1] + adj[i-1]
        new_x = max(x_sort[i], prev + local_dmin)        
        slack[i] = new_x - (prev + local_dmin)
        adj[i] = new_x - x_sort[i]

    new_x = x_sort.copy()
    chains = np.ones_like(new_x)
    
    for i in range(1, len(new_x)):
        if slack[i] == 0:
            chains[i] = chains[i - 1]
        else:
            chains[i] = chains[i - 1] + 1

    # We break the elements into "chains", groups that are all the minimum distance apart.
    # To start out, we make chains as above, simply checking whether each element is tied
    # to the one before it (because that can be easily computed)

    # This produces incorrect results when some elements move left, thereby entangling more.
    # For example, with dmin = 3 the list [0, 4, 5, 6] is actually all connected, because
    # the chain [4, 5, 6] will become [2, 5, 8] which is within range of 0. 

    # To adjust a single chain, we compute the median of the shifts if only rightward shifts
    # are allowed, and subtract that median from each of the shifts.

    # We repeat the following process until convergence:
    # - Adjust each chain independently
    # - If two chains are within the minimum distance of each other, merge them

    def adjust(chains):
        for c in np.unique(chains):
            chain_adj = adj[chains == c]
            chain_adj -= np.max(chain_adj) / 2
            new_x = x_sort.copy()
            new_x[chains == c] += chain_adj
        return new_x

    def chain_overlaps(chains):
        adj_x = adjust(chains)
        intervals = []
        for c in np.unique(chains):
            inds = np.nonzero(chains == c)[0]            
            intervals.append((
                adj_x[inds[0]] - dmin[inds[0]],
                adj_x[inds[0]], 
                adj_x[inds[-1]],
                adj_x[inds[-1]] + dmin[inds[-1]]))

        # print(intervals)
        overlaps = [False]
        for i in range(1, len(intervals)):
            (lo1, a1, b1, hi1), (lo2, a2, b2, hi2) = intervals[i-1], intervals[i]            
            overlaps.append(b1 > lo2 or a2 < hi1)

        return np.array(overlaps)

    overlaps = chain_overlaps(chains)
    while overlaps.any():    
        # print(chains)
        # print(adjust(chains))
        # print(overlaps)
    
        for i in range(1, len(overlaps)):
            if overlaps[i]:
                print(i, overlaps)
                left, right = np.unique(chains)[i-1:i+1]
                chains[chains == right] = left

        overlaps = chain_overlaps(chains)

    return adjust(chains)[np.argsort(sort_inds)]


def labelcase(text: Union[str, Iterable[str]]):
    """Converts text to a title case and adds spacing from camel case."""
    if isinstance(text, str):
        if "_" in text:
            # snake_case
            text = text.replace("_", " ")
        return re.sub(r"([^ ])([A-Z])", r"\1 \2", text).title().strip()
    else:
        return [labelcase(t) for t in text]



from .matplotlib import setup as mpl_setup
from .matplotlib import boxstyle as boxstyle
from .matplotlib import unfill_boxplot as unfill_boxplot
from .matplotlib_tweaks import smart_ticks, line_labels, ylabel_top

try:
    import altair
    from .vega import setup as vega_setup
    from .vega import RHO_LIGHT as vega_rho_light
    from .vega import RHO_DARK as vega_rho_dark
    from .vega import _register_themes

    _register_themes()
except ModuleNotFoundError:
    pass

try:
    import bokeh

    from .bokeh import setup as bokeh_setup
    from .bokeh import rho_light as bokeh_rho_light
    from .bokeh import rho_dark as bokeh_rho_dark
except ModuleNotFoundError:
    pass

try:
    from .panel import ThemedPanel
    from .panel import pn_setup_fonts

    pn_setup_fonts()
except ModuleNotFoundError:
    pass

try:
    import plotly
    from .plotly import setup as plotly_setup
    from .plotly import register_themes as _register_themes

    _register_themes()

except ModuleNotFoundError:
    pass

from .sequential_palettes import *



#!/usr/bin/env python3
"""Plotly theming."""

import json

import matplotlib as mpl

import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio

import plotly.graph_objects as go
import numpy as np

from .matplotlib import setup as mpl_setup

from .sequential_palettes import SEQUENTIAL


def register_themes():
    for theme_name, is_dark in zip(("rho_light", "rho_dark"), (False, True)):
        inherit = "plotly_dark" if is_dark else "plotly"
        default = pio.templates[inherit].layout

        templ: go.Layout = go.Layout(default)
        templ.colorscale.sequential = templ.colorscale.sequentialminus = SEQUENTIAL[
            "viridia"
        ].hex_colors()[::17]
        templ.colorscale.diverging = (
            SEQUENTIAL["div_icefire_shift" if is_dark else "div_coolwarm_shift"]
        ).hex_colors()[2:-1:18]

        _theme, cs = mpl_setup(is_dark=is_dark, setup=True)
        rc = mpl.rcParams

        templ.paper_bgcolor = rc["axes.facecolor"]
        templ.plot_bgcolor = rc["axes.facecolor"]
        templ.colorway = cs
        templ.margin = go.layout.Margin(b=20, l=20, r=20, t=30)
        for ax in (
            templ.xaxis,
            templ.yaxis,
            templ.scene.xaxis,
            templ.scene.yaxis,
            templ.scene.zaxis,
        ):
            ax.tickcolor = rc["xtick.color"]
            ax.linecolor = rc["axes.edgecolor"]

            ax.gridcolor = rc["axes.edgecolor"]
            ax.tickfont["color"] = rc["xtick.labelcolor"]
            ax.title.font["color"] = rc["axes.labelcolor"]
            ax.showline = True
            ax.showgrid = False
            ax.mirror = False

        for ax in (templ.scene.xaxis, templ.scene.yaxis, templ.scene.zaxis):
            ax.showgrid = True
            ax.backgroundcolor = rc["axes.facecolor"]

        templ.legend.bgcolor = rc["axes.facecolor"]
        templ.legend.bordercolor = rc["legend.edgecolor"]
        templ.legend.borderwidth = 2
        templ.legend.font.color = rc["legend.labelcolor"]
        templ.legend.title.font.color = rc["axes.labelcolor"]

        # I'm tempted to keep Plotly's default of using the color of the mark for the hover color, but ultimately
        # I think this is more minimal without being any harder to understand: if you hover, you're already
        # looking at the data, so the extra color cue isn't especially necessary
        templ.hoverlabel.bgcolor = rc["axes.facecolor"]
        templ.hoverlabel.bordercolor = rc["axes.edgecolor"]
        templ.hoverlabel.font.color = rc["text.color"]

        data = go.layout.template.Data(
            # https://github.com/plotly/plotly.py/issues/3404
            # I'd rather have filled box plots by default, but Plotly doesn't support altering the default of half-transparency fill color (God only knows why...)
            # so I can't make the fill color just the line color, so my choices are half transparency or transparent
            box=(
                go.Box(
                    marker_line_color=rc["axes.facecolor"],
                    notchwidth=0,
                    fillcolor="rgba(0,0,0,0)",
                ),
            ),
            # same issue as box plots with violin plots: here, we have the fill because otherwise it looks a little too empty
            # I'd rather have a solid meanline, but that's not an option
            violin=(go.Violin(meanline=dict(visible=True)),),
            scatter=(go.Scatter(marker=dict(line=dict(color=rc["axes.facecolor"]))),),
        )
        pio.templates[theme_name] = go.layout.Template(layout=templ, data=data)


def setup(is_dark):
    pio.templates.default = "rho_dark" if is_dark else "rho_light"



from __future__ import annotations

"""Uses Oklch as a color space to programmatically generate perceptually uniform color palettes.
To keep the import cost of Rho+ small, the modules necessary to do this are not part of the package."""

import abc
from typing import List
import matplotlib.colors as mpl_colors
import numpy as np
from scipy import interpolate as interp
import colour
from colour.utilities import suppress_warnings
from scipy import stats as stats
import scipy.signal as signal
from .palettes import SequentialPaletteMixin


def hex2oklch(hexs):
    with suppress_warnings(colour_usage_warnings=True):
        lchs = []
        if isinstance(hexs, str):
            hexs = [hexs]
        for hex in hexs:
            L, a, b = colour.convert(mpl_colors.to_rgb(hex), "sRGB", "Oklab")
            h = np.rad2deg(np.arctan2(b, a)) % 360
            c = np.hypot(b, a)
            lchs.append(np.array([L, c, h]))

    return np.array(lchs)


def oklch2hex(oklch):
    with suppress_warnings(colour_usage_warnings=True):
        return np.array(
            [
                mpl_colors.to_hex(
                    np.clip(
                        colour.convert(
                            (L, c * np.cos(np.deg2rad(h)), c * np.sin(np.deg2rad(h))),
                            "Oklab",
                            "sRGB",
                        ),
                        0,
                        1,
                    )
                )
                for L, c, h in np.array(oklch).reshape(-1, 3)
            ]
        )


def hex2oklab(hexs):
    with suppress_warnings(colour_usage_warnings=True):
        labs = []
        if isinstance(hexs, str):
            hexs = [hexs]
        for hex in hexs:
            L, a, b = colour.convert(mpl_colors.to_rgb(hex), "sRGB", "Oklab")
            labs.append(np.array([L, a, b]))

        return np.array(labs)


def oklab2hex(oklab):
    with suppress_warnings(colour_usage_warnings=True):
        return np.array(
            [
                mpl_colors.to_hex(np.clip(colour.convert(lab, "Oklab", "sRGB"), 0, 1))
                for lab in np.array(oklab).reshape(-1, 3)
            ]
        )


def oklch2scd(oklch):
    with colour.utilities.suppress_warnings(colour_usage_warnings=True):
        return np.array(
            [
                colour.convert(
                    (L, c * np.cos(np.deg2rad(h)), c * np.sin(np.deg2rad(h))),
                    "Oklab",
                    "CAM16SCD",
                )
                for L, c, h in np.array(oklch).reshape(-1, 3)
            ]
        )


def oklab2oklch(labs):
    lchs = []
    for (L, a, b) in labs.reshape(-1, 3):
        h = np.rad2deg(np.arctan2(b, a)) % 360
        c = np.hypot(b, a)
        lchs.append(np.array([L, c, h]))

    return np.array(lchs)


def oklch2oklab(lchs):
    labs = []
    for (L, c, h) in lchs.reshape(-1, 3):
        a = c * np.cos(np.deg2rad(h))
        b = c * np.sin(np.deg2rad(h))
        labs.append(np.array([L, a, b]))

    return np.array(labs)


class OklchPalette(SequentialPaletteMixin):
    """A sequential color palette in Oklch space."""

    def __init__(self, name, keys, hue_cool, nodes=None, correct=True):
        """Keys is a list of colors. Hue_cool is a matching list of Booleans (with first element removed)
        indicating whether the interpolation to that color should go forward (ROYGBIV) or backwards.
        Nodes, if given, is a list of numbers the same length as keys that indicate how they should be spaced.
        (Scaled within [0, 1].)
        If correct, then apply a correction to ensure perceptual uniformity."""

        if isinstance(keys[0], str):
            self.keys = keys
            self.lch_keys = hex2oklch(keys)
        else:
            self.keys = oklch2hex(keys)
            self.lch_keys = [list(key) for key in keys]
        self._name = name

        for i, cool in enumerate(hue_cool):
            # if equal or very close, do nothing
            if abs(self.lch_keys[i][2] - self.lch_keys[i + 1][2]) < 1e-2:
                continue

            # whether the hue will already go forwards
            cool_default = self.lch_keys[i][2] < self.lch_keys[i + 1][2]
            if cool == cool_default:
                # no need to do anything
                pass
            elif cool and not cool_default:
                # hue 2 is smaller than hue 1, but it needs to wrap around the other way
                # add 360 to hue 2
                self.lch_keys[i + 1][2] += 360
            elif not cool and cool_default:
                # hue 2 is larger, but needs to wrap around the other way
                # subtract 360 from second hue
                self.lch_keys[i + 1][2] -= 360

        if nodes is None:
            self.nodes = np.linspace(0, 1, len(keys))
        else:
            self.nodes = (np.array(nodes) - np.min(nodes)) / np.ptp(nodes)

        self.spline = interp.interp1d(self.nodes, self.lch_keys, kind="linear", axis=0)
        self.uncorrected_spline = self.spline
        if not correct:
            pass
        elif correct == "cam16":
            self.cam16_correct()
        elif correct == "gauss":
            self.gaussian_correct()
        else:
            self.correct()

    def correct(self):
        """Corrects for any perceptual uniformity issues by resampling the colormap."""
        # https://arxiv.org/pdf/1509.03700.pdf
        # at small scales, lightness is the only major factor in perceptual difference
        # so we get the cumulative lightness changes and then resample to evenly space those out

        # cumulative lightness change (now monotonic)
        ll = hex2oklch(self.colors(256))[:, 0]
        abs_ll = np.cumsum(np.abs(np.diff(ll, prepend=0)))

        # interpolate from % of total cumulative lightness to [0, 1] in original
        # only do this if the color scheme has meaningful lightness change
        if np.ptp(abs_ll) > 0.05:
            corr_spline = interp.interp1d(
                (abs_ll - np.min(abs_ll)) / np.ptp(abs_ll),
                np.linspace(0, 1, len(abs_ll)),
                kind="linear",
                axis=0,
            )

            # this spline applied to an evenly-spaced list produces the x
            # required to achieve that equispaced map in the original
            self.spline = lambda x: self.uncorrected_spline(corr_spline(x))

    def cam16_correct(self):
        """Corrects for any perceptual uniformity issues by resampling the colormap for CAM16 uniformity."""
        lch = hex2oklch(self.colors(256))
        scd = oklch2scd(lch)
        diffs = np.diff(scd, prepend=0, axis=0)
        abs_ll = np.cumsum(np.linalg.norm(diffs, axis=1))

        # interpolate from % of total cumulative lightness to [0, 1] in original
        # only do this if the color scheme has meaningful lightness change
        if np.ptp(abs_ll) > 0.05:
            corr_spline = interp.interp1d(
                (abs_ll - np.min(abs_ll)) / np.ptp(abs_ll),
                np.linspace(0, 1, len(abs_ll)),
                kind="linear",
                axis=0,
            )

            # this spline applied to an evenly-spaced list produces the x
            # required to achieve that equispaced map in the original
            self.spline = lambda x: self.uncorrected_spline(corr_spline(x))

    def gaussian_correct(self):
        """Uses a Gaussian filter to smooth the variation, removing discontinuities."""
        # https://arxiv.org/pdf/1509.03700.pdf
        # the more smoothing, the more the area in the middle is difficult to distinguish, but the less it pops out as a false feature
        # sigma around 5-7 for a 256-color map is recommended

        # cumulative lightness change (now monotonic)
        lab = hex2oklab(self.colors(256))
        # pad with the end values
        window_size = 40
        lab_padded = np.vstack(
            [
                np.tile(lab[0], (window_size // 2, 1)),
                lab,
                np.tile(lab[-1], (window_size // 2, 1)),
            ]
        )
        window = stats.norm.pdf(
            np.arange(-window_size // 2, window_size // 2 + 1), scale=6
        )
        window /= np.sum(window)
        smoothed_lab = np.vstack(
            [signal.convolve(lab_padded[:, i], window, mode="same") for i in range(3)]
        ).T
        smoothed_lab = smoothed_lab[window_size // 2 : -window_size // 2]
        self.spline = interp.interp1d(
            np.linspace(0, 1, len(lab)),
            oklab2oklch(smoothed_lab),
            kind="linear",
            axis=0,
        )

    def colors(self, n):
        """Gets the palette as a list of n colors."""
        return oklch2hex(self.spline(np.linspace(0, 1, n)))

    def hex_colors(self):
        return self.colors(256)

    def name(self):
        return self._name



"""Matplotlib tweaks to plots that enable more complicated customizations from Matplotlib."""

import textwrap

import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt
from .util import spread as rho_spread


def remove_crowded(ax=None):
    """Given an axis, removes the second and penultimate ticks if
    the distance between the final ticks and those are less than half of the
    average distance."""
    if ax is None:
        ax = plt.gca()

    # check in view coordinates to adjust for log axes and the like
    ticks = [ax.get_xticks(), ax.get_yticks()]
    xticklocs = ax.transData.transform(
        np.vstack([ticks[0], np.ones_like(ticks[0]) * ticks[1][0]]).T
    )[:, 0]
    yticklocs = ax.transData.transform(
        np.vstack([np.ones_like(ticks[1]) * ticks[0][0], ticks[1]]).T
    )[:, 1]

    tick_locs_list = [xticklocs, yticklocs]
    for i, tick_locs in enumerate(tick_locs_list):
        meandist = np.diff(tick_locs).mean()
        if tick_locs[1] - tick_locs[0] < 0.5 * meandist:
            ticks[i] = np.array([ticks[i][0], *ticks[i][2:]])

        if tick_locs[-1] - tick_locs[-2] < 0.5 * meandist:
            ticks[i] = np.array([*ticks[i][:-2], ticks[i][-1]])

    ax.set_xticks(ticks[0])
    ax.set_yticks(ticks[1])


def smart_ticks(ax=None, xaxis=True, yaxis=True):
    """Sets the axis limits so they start and end at the minimum and maximum for each axis, a la Tufte."""
    if ax is None:
        ax = plt.gca()

    bboxes = []
    for child in ax.get_children():
        if isinstance(child, mpl.collections.Collection):
            bbox = child.get_datalim(ax.transData)
        elif isinstance(child, mpl.lines.Line2D):
            xdata, ydata = child.get_data()
            if len(xdata):
                bbox = mpl.transforms.Bbox.from_extents(
                    xdata.min(), ydata.min(), xdata.max(), ydata.max()
                )
            else:
                continue
        else:
            continue
        bboxes.append(bbox)

    if not len(bboxes):
        return None

    bbox = mpl.transforms.Bbox.union(bboxes)
    ax.spines["bottom"].set_bounds(bbox.xmin, bbox.xmax)
    ax.spines["left"].set_bounds(bbox.ymin, bbox.ymax)

    # ignore fixed locators, which are often not numeric
    if xaxis and not isinstance(ax.xaxis.get_major_locator(), mpl.ticker.FixedLocator):
        xlabels = ax.xaxis.get_major_formatter().format_ticks([bbox.xmin, bbox.xmax])
        xlabels = [
            xlabels[0],
            *ax.xaxis.get_major_formatter().format_ticks(ax.get_xticks()[1:-1]),
            xlabels[-1],
        ]
        xticks = [bbox.xmin, *ax.get_xticks()[1:-1], bbox.xmax]
        ax.set_xticks(xticks, xlabels)

    if yaxis and not isinstance(ax.yaxis.get_major_locator(), mpl.ticker.FixedLocator):
        ylabels = ax.yaxis.get_major_formatter().format_ticks([bbox.ymin, bbox.ymax])
        ylabels = [
            ylabels[0],
            *ax.yaxis.get_major_formatter().format_ticks(ax.get_yticks()[1:-1]),
            ylabels[-1],
        ]
        yticks = [bbox.ymin, *ax.get_yticks()[1:-1], bbox.ymax]
        ax.set_yticks(yticks, ylabels)

    remove_crowded(ax)


def get_handles_labels(ax):
    """Gets a tuple (handles, labels). Accounts for Seaborn's use of phantom lines, so the lines
    will actually have data on the screen."""
    all_handles, labels = ax.get_legend_handles_labels()
    if len(ax.get_lines()) == 2 * len(all_handles):
        # Seaborn does line plots with a legend by first plotting all of the line data,
        # then making another list that same size of phantom lines without data
        handles = ax.get_lines()[: len(ax.get_lines()) // 2]
    else:
        # none of that weirdness
        handles = all_handles

    return handles, labels


def height(text, ax, **kwargs):
    t = ax.text(0.5, 0.5, text, transform=ax.transAxes, **kwargs)
    ax.figure.canvas.draw()
    bb = t.get_window_extent()
    t.remove()
    return bb.height


def line_labels(ax=None, remove_legend=True, spread=None):
    """Does automatic line labeling, replacing a legend with side labels."""
    if ax is None:
        ax = plt.gca()

    if spread is None:
        spread = rho_spread

    handles, labels = get_handles_labels(ax)
    labels = ["\n".join(textwrap.wrap(label, width=15)) for label in labels]
    data_to_in = mpl.transforms.CompositeGenericTransform(
        ax.transData, ax.figure.dpi_scale_trans.inverted()
    )

    margin = np.array([height(label, ax) for label in labels])

    y_ends = np.array([handle.get_xydata()[-1] for handle in handles])

    spreaded_y_ends = ax.transData.transform(y_ends)
    spreaded_y_ends[:, 1] = spread(spreaded_y_ends[:, 1], margin)
    spreaded_y_ends = ax.transData.inverted().transform(spreaded_y_ends)

    xmax = ax.transData.inverted().transform(
        (ax.transData.transform((y_ends[:, 0].max(), 0))[0] * 1.05, 0)
    )[0]
    for y_end, spreaded, handle, label in zip(y_ends, spreaded_y_ends, handles, labels):
        color = handle.get_color()
        ax.plot([y_end[0], xmax], [y_end[1], spreaded[1]], ls="--", lw=1, c=color)

        ax.text(xmax, spreaded[1], label, ha="left", va="center", color=color)

    if remove_legend:
        ax.legend().remove()


def ylabel_top(ax=None):
    """Moves the y-axis label to the top, wrapping as necessary."""
    if ax is None:
        ax = plt.gca()

    texts = [
        text
        for text in ax.yaxis.get_children()
        if isinstance(text, mpl.text.Text) and text.get_text() == ax.get_ylabel()
    ]

    text = texts[0]
    text.set_visible(False)

    label_wrap = "\n".join(textwrap.wrap(ax.get_ylabel(), width=15))
    ax.text(
        0,
        1.03,
        label_wrap,
        transform=ax.transAxes,
        ha="right",
        va="baseline",
        color=text.get_color(),
        fontproperties=text.get_font_properties(),
    )



#!/usr/bin/env python3
from .bokeh import rho_light, rho_dark
from .matplotlib import setup as mpl_setup
import param
import panel as pn


def pn_setup_fonts():
    try:
        import panel as pn

        pn.config.raw_css.append(
            """
@import url('https://fonts.googleapis.com/css2?family=Source+Sans+3:ital,wght@0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&display=swap');
"""
        )
    except ModuleNotFoundError:
        pass


class ThemedPanel(param.Parameterized):
    colorMode = param.ObjectSelector("dark", ["dark", "light"], precedence=-1)

    def __init__(self):
        super().__init__()
        if "colorMode" in pn.state.session_args:
            self.colorMode = pn.state.session_args["colorMode"][0].decode().lower()
        else:
            try:
                self.colorMode = "dark" if IS_DARK else "light"
            except NameError:
                self.colorMode = "dark"

    @pn.depends("colorMode")
    def colors_theme(self):
        if self.colorMode == "light":
            _theme, colors = mpl_setup(False, False)
        else:
            _theme, colors = mpl_setup(True, False)

        theme = rho_light if self.colorMode == "light" else rho_dark
        return (colors, theme)



#!/usr/bin/env python3
"""Vega plot themes as config objects."""

import altair as alt

from .colors import LIGHT_SHADES, DARK_SHADES, LIGHT_COLORS, DARK_COLORS
from .matplotlib import rho_dark, rho_light
from .matplotlib import setup as mpl_setup
from .sequential_palettes import SEQUENTIAL
import matplotlib.colors


def color_values(theme):
    # more values is closer to the original LCH space, but more data
    return [
        matplotlib.colors.rgb2hex(x)
        for x in theme.as_mpl_cmap()([i / 15 for i in range(16)])
    ]


# define the theme by returning the dictionary of configurations
def rho(is_dark: bool):
    theme, colors = mpl_setup(is_dark, False)
    empty, lightest, light, medium, dark, darkest = (
        DARK_SHADES if is_dark else LIGHT_SHADES
    )

    def r():
        return {
            "config": {
                "background": "#" + theme["figure.facecolor"],
                "numberFormat": ".5~r",
                "range": {
                    "category": colors,
                    "heatmap": color_values(SEQUENTIAL["gouldia"]),
                    "diverging": color_values(
                        SEQUENTIAL[
                            "div_icefire_shift" if is_dark else "div_coolwarm_shift"
                        ]
                    ),
                    "ramp": color_values(SEQUENTIAL["viridia"]),
                },
                "circle": {"size": 100, "fill": colors[0],},
                "square": {"size": 100, "fill": colors[0],},
                "line": {"strokeWidth": 3,},
                "style": {
                    "guide-label": {"fill": dark, "fontWeight": 400, "fontSize": 16,},
                    "guide-title": {
                        "fill": darkest,
                        "fontSize": 20,
                        "fontWeight": 400,
                    },
                    "group-title": {
                        "fill": darkest,
                        "fontSize": 22,
                        "fontWeight": 700,
                    },
                },
                "axis": {
                    "tickFontWeight": 600,
                    "tickColor": dark,
                    "domainColor": dark,
                },
                "axisY": {"titleAngle": 0, "titleAlign": "right"},
                "axisXBand": {"labelAngle": -45},
                "axisQuantitative": {"grid": False},
                "legend": {
                    "labelColor": darkest,
                    "gradientHorizontalMinLength": 200,
                    "gradientHorizontalMaxLength": 1000,
                    "gradientVerticalMinLength": 200,
                    "gradientVerticalMaxLength": 1000,
                },
                "view": {
                    "background": "#" + theme["axes.facecolor"],
                    "stroke": "transparent",
                },
            }
        }

    return r


RHO_LIGHT = rho(False)()
RHO_DARK = rho(True)()


def _register_themes():
    alt.themes.register("rho_dark", rho(True))
    alt.themes.register("rho_light", rho(False))


def setup(is_dark: bool):
    """Sets up Altair according to the given color scheme."""
    alt.themes.enable("rho_dark" if is_dark else "rho_light")



#!/usr/bin/env python3
"""Bokeh theme."""

import rho_plus
import matplotlib.pyplot as plt
from bokeh.themes import Theme

themes = []
for is_dark in (True, False):
    _theme, cs = rho_plus.mpl_setup(is_dark)
    theme = plt.rcParams

    font = "'Source Sans 3', sans-serif"
    json = {
        "attrs": {
            "Figure": {
                "background_fill_color": theme["figure.facecolor"],
                "border_fill_color": theme["figure.facecolor"],
                # spine of axis
                "outline_line_color": theme["axes.facecolor"],
            },
            "Grid": {"grid_line_alpha": 0, "minor_grid_line_alpha": 0,},
            "Title": {
                "text_color": theme["axes.titlecolor"],
                "text_font": font,
                "align": "center",
                "text_font_size": "1.4em",
                "text_font_style": "normal",
            },
            "Axis": {
                "major_tick_line_color": theme["xtick.color"],
                "major_tick_line_width": 1,
                "major_tick_out": 10,
                "minor_tick_line_color": theme["xtick.color"],
                "major_label_text_color": theme["xtick.labelcolor"],
                "major_label_text_font": font,
                "major_label_text_font_size": "1.025em",
                "minor_tick_out": 0,
                "axis_label_standoff": 10,
                "axis_label_text_color": theme["axes.labelcolor"],
                "axis_label_text_font": font,
                "axis_label_text_font_size": "1.25em",
                "axis_label_text_font_style": "normal",
            },
            "YAxis": {"major_label_orientation": "horizontal"},
            "Legend": {
                "spacing": 8,
                "glyph_width": 15,
                "label_standoff": 8,
                "label_text_color": theme["legend.labelcolor"],
                "label_text_font_size": "1.025em",
                "border_line_alpha": 1,
                "border_line_color": theme["legend.edgecolor"],
                "background_fill_alpha": 0.25,
                "background_fill_color": theme["axes.facecolor"],
            },
            "BaseColorBar": {
                "title_text_color": theme["axes.titlecolor"],
                "title_text_font": font,
                "title_text_font_size": "1.025em",
                "major_label_text_color": theme["axes.titlecolor"],
                "major_label_text_font": font,
                "major_label_text_font_size": "1.025em",
                "background_fill_color": theme["axes.facecolor"],
                "major_tick_line_alpha": 0,
                "bar_line_alpha": 0,
            },
            "Line": {"line_width": 4,},
        }
    }

    themes.append(Theme(json=json))

rho_dark, rho_light = themes


def setup(is_dark):
    theme = rho_dark if is_dark else rho_light
    try:
        import holoviews as hv

        hv.renderer("bokeh").theme = theme
    except ModuleNotFoundError:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rho_plus', 'rho_plus.data']

package_data = \
{'': ['*'], 'rho_plus.data': ['fonts/*']}

install_requires = \
['scipy>=1.7.0']

setup_kwargs = {
    'name': 'rho-plus',
    'version': '0.4.5',
    'description': 'Aesthetic and ergonomic enhancements to common Python data science tools',
    'long_description': None,
    'author': 'Nicholas Miklaucic',
    'author_email': 'nicholas.miklaucic@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
